
import numpy as np
import pandas as pd

import argparse
import os, glob
from sys import argv

##=================================================================================================

me = argv[0]

##=================================================================================================
##=================================================================================================


def main():
	"""
	Clean Eurostat agricultural productivity data and output a formatted csv of yields.
    
    Main input: Eurostat dataset apro_cpshr
    Main output: yield dataset, at the NUTS2 level and pivoted to each year
    
    Other inputs: crop IDs, NUTS regions
    Other outputs: diagnostic files from the cleaning
    
    Cautions: The run will fail if the directory structure is not set up properly.
    
    Execution:  python Eurostat_DistrictYield_Clean [options and flags]
    
    Please feel free to contact cato@cerulogy.com with questions or improvements :)
	"""
	
	##---------------------------------------------------------------------------------------------
	## CLAs
	
	parser = argparse.ArgumentParser()
	parser.add_argument("-d", "--datadir", type=str, default="./Data", 
		help="Directory of the raw datafile.")
	parser.add_argument("-r", "--rawfile", type=str, default="apro_cpshr_linear.csv", 
		help="The filename for the original Eurostat data download.")
	parser.add_argument("-f", "--force", default=False, action="store_true",
		help="Force re-cleaning of the raw dataset.")
	parser.add_argument(      "--nosave", default=False, action="store_true", 
		help="Toggle file output off.")
	parser.add_argument("-q", "--quiet", default=False, action="store_true", 
		help="Minimise stdout info.")
	args = parser.parse_args()
	
	## Name args
	datadir = args.datadir
	force = args.force
	nosave = args.nosave
	verbose = not args.quiet
	
	##---------------------------------------------------------------------------------------------
	## Define filenames
    
    ## Identify source file
	datafile_raw = datadir + "/" + args.rawfile
    
    ## Checks
	assert os.path.isdir(datadir), me+": Check input directory."
	assert os.path.isfile(datafile_raw), me+": Input file not found."
	
    ## Names of output files
	datafile = os.path.splitext(datafile_raw)[0]+"_cleaned.csv"
	pivotfile = os.path.splitext(datafile)[0]+"_pivot.csv"
	yieldfile = os.path.splitext(datafile)[0]+"_yield.csv"
		
	##---------------------------------------------------------------------------------------------
	## Read in cleaned data, or clean raw data
	
	## Read big cleaned datafile (overhead, but ensures simplicity)
	if os.path.isfile(datafile) and not force:
		data = pd.read_csv(datafile, header=0, index_col=0)
		if verbose: print(me+": Have read in big cleaned datafile {0}.".format(datafile))
	
	## Read big raw datafile, clean, and save
	elif os.path.isfile(datafile_raw):
		if verbose: print(me+": Cleaned file not found. Trying raw file {0}.".format(datafile_raw))
		data = read_raw_data(datafile_raw, verbose, nosave)
	
	## Files not found
	else:
		raise IOError(me+": Data files not found: check input directory."+
			"\n\tClean: {0}\n\tRaw: {1}".format(datafile,datafile_raw))
		
	##---------------------------------------------------------------------------------------------
	## Create a pivot dataset

	## Keep only area and production data
	## This also drops existing yield data (will be re-calculated)
	idx = (data["Variable ID"]=="AR") | (data["Variable ID"]=="PR_HU_EU")
	data = data.loc[idx]

	## Clean zeroes from the raw dataset
	## No need to preserve this information and it cuases problems later
	data.replace([np.inf,-np.inf,0], np.nan, inplace=True)
	
	## Pivot the data to annual columns
	data_pivot = clean_dopivoting(data)
	data_pivot.replace([np.inf,-np.inf,0], np.nan, inplace=True)
	
	##---------------------------------------------------------------------------------------------
	## Check for duplicated data
	
	## In raw datafile
	data = clean_duplicates_raw(data, datafile, verbose, nosave)
	
	## In pivoted data
	data_pivot = clean_duplicates_pivot(data_pivot, datafile, verbose, nosave)

	##---------------------------------------------------------------------------------------------
	## Check for quirks
	
	## Check if there are suspicious whole numbers and save a report (df doesn't change)
	check_rounding(data_pivot, datafile, verbose, nosave)
	
	##---------------------------------------------------------------------------------------------
	## Output
	
	if not nosave:
		data_pivot.to_csv(pivotfile)
		if verbose: print(me+": Pivoted and cleaned data saved to {outfile:s}".format(outfile=pivotfile))
		
	##---------------------------------------------------------------------------------------------
	##---------------------------------------------------------------------------------------------
	## Yield calculation and cleaning
	
	## Calculate yield from input data and clean up
	data_yield = (data_pivot.loc[("PR_HU_EU"),("Value")]/data_pivot.loc[("AR"),("Value")]).round(decimals=2)
	data_yield.replace([np.inf,-np.inf, 0], np.nan, inplace=True)
	if verbose: print(me+": Calculated yields; now cleaning.".format(outfile=pivotfile))
	
	## Clean yield data -- merge NUTS2 regions
	data_yield = clean_NUTSupdate_merge(data_yield, verbose, nosave)
	
	## Clean yield data -- identify jumps
	data_yield = clean_yielddata_maskjumps(data_yield, verbose, nosave)
	
	## Clean yield data -- fill crop data from parent crop if missing
	data_yield = clean_yielddata_fillcrops(data_yield, verbose)
	
	##---------------------------------------------------------------------------------------------
	## Output
	
	if not nosave:
		data_yield.to_csv(yieldfile)
		if verbose: print(me+": Yield data saved to {outfile:s}".format(outfile=yieldfile))
	
	return

##=================================================================================================
## FILE READ
##=================================================================================================

def read_raw_data(datafile, verbose, nosave):
	"""
	Read in raw Eurostat data file and return DataFrame
	Standardise column names
	"""
	
	## Read file
	data = pd.read_csv(datafile, header=0, index_col=None)
	print(me+": Read in big raw datafile {0}.".format(datafile))
	
	##---------------------------------------------------------------------------------------------
	## Simple cleaning
	
	## Drop unnecessary columns
	for col in ["DATAFLOW","LAST UPDATE","freq","OBS_FLAG"]:
		try:
			data.drop(col, axis="columns", inplace=True)
		except KeyError:
			continue
	
	## Rename columns
	data.rename(columns=
		{"crops":"Crop ID",
		"strucpro":"Variable ID",
		"geo":"Area ID",
		"TIME_PERIOD":"Year",
		"OBS_VALUE":"Value"}, inplace=True)
	
	##---------------------------------------------------------------------------------------------
	## Update NUTS2 names where there is simple recoding
	
	data = clean_NUTSupdate_recode(data.copy(), verbose, nosave)
	
	##---------------------------------------------------------------------------------------------
	## Add full names to standard IDs
	
	for IDmap in [read_NUTSID(),read_VariableID(),read_CropID()]:
		data = clean_IDToName(data, IDmap)
	
	data.drop(data.columns[data.columns.str.contains("unnamed",case=False)], axis=1, inplace=True)
	
	##---------------------------------------------------------------------------------------------
	## Save cleaned datafile
	
	if not nosave:
		outfile = os.path.splitext(datafile)[0]+"_cleaned.csv"
		data.to_csv(outfile)
		if verbose: print(me+": Cleaned data saved to "+outfile)
	
	return data

##=================================================================================================

def read_NUTSID():
	"""
	Read in the NUTS lookup table, in order to assign names to area IDs
	Outputs a DataFrame, indexed by NUTS ID
	"""
	
	## Identify where the area codes are kept
	dictfile = "./CodesLookup/NUTS2021.xlsx"
	data_lookup_raw = pd.read_excel(dictfile, sheet_name="NUTS & SR 2021", header=0, index_col=0, 
		usecols=["Code 2021","Country","NUTS level 1","NUTS level 2","NUTS level 3","NUTS level","Country order","Region order"])
	data_lookup_raw.index.name = "Area ID"
	
	## Clean out empty rows to create a contiguous dataset
	data_lookup_raw.dropna(axis="index", how="all", inplace=True)
	data_lookup_raw.fillna("", inplace=True)
	
	## Create new DataFrame to act as an easy index-lookup
	data_lookup = pd.DataFrame(index=data_lookup_raw.index, columns=["Area Name","NUTS Level"])
	data_lookup["NUTS Level"] = data_lookup_raw["NUTS level"].astype("int")
	
	## Concatenate all area designators (only one will have a value)
	data_lookup["Area Name"] = data_lookup_raw[["Country","NUTS level 1","NUTS level 2","NUTS level 3"]].astype(str).apply("".join, axis=1)
	
	## Drop multiple names for brevity
	data_lookup["Area Name"] = data_lookup["Area Name"].str.split(pat="/", expand=False).str[0]
	
	return data_lookup

##=================================================================================================

def read_CountryID():
	"""
    Return a dataframe of country IDs and names
	"""
	
	## Identify where the area codes are kept and read in
	dictfile = "./CodesLookup/CountryCodes.csv"
	data_lookup = pd.read_csv(dictfile, header=0, index_col=0)
		
	return data_lookup
	
##=================================================================================================

def read_VariableID():
	"""
	Manually construct a map between variable abbreviations and long names
	No file is read for this routine
	Returns a DataFrame, indexed by variable ID
	"""
	
	## Define map
	variableID_list = ["MA","AR","HU_EU","PR_HU_EU","YI_HU_EU"]
	variablename_list = ["Main Area (ha)","Harvested Area (ha)","Humidity (%)", "Production (t)", "Yield (t/ha)"]
	
	## Create dataframe
	data_lookup = pd.DataFrame(index=variableID_list, columns=["Variable Name"], data=variablename_list)
	data_lookup.index.name = "Variable ID"
	
	return data_lookup
	
##=================================================================================================

def read_CropID():
	"""
	Read in the Eurostat crop ID lookup table, in order to assign names
	Returns a DataFrame, indexed by crop ID
	"""
	
	## Identify where the area codes are kept and read in
	dictfile = "./CodesLookup/CL_AGRIPRO_20220509_154326.csv"
	data_lookup = pd.read_csv(dictfile, header=0, index_col=None, usecols=["Code.1","Description","Parent.1"])
		
	## Set index to ID and rename headers
	data_lookup.rename(columns=
		{"Code.1":"Crop ID", "Description":"Crop Name", "Parent.1":"Parent ID"},
		inplace=True)
	data_lookup.set_index("Crop ID", drop=True, inplace=True)
	data_lookup.index.name = "Crop ID"
	
	## Truncate long-name strings
	data_lookup["Crop Name"] = data_lookup["Crop Name"].apply(lambda x: x if len(x)<20 else x[:18]+"...")
	
	return data_lookup

##=================================================================================================
## DATA CLEANING
##=================================================================================================

def clean_NUTSupdate_recode(data, verbose, nosave):
	"""
	Replace old NUTS2 area codes with new ones
	This is the first (and simpler) part of a two-stage process of area code updates
	It operates on the raw pre-pivot data and simply updates the codes
	"""
	
	## Read in file of ID changes
	NUTSfile = "./CodesLookup/NUTS2_ChangeCompilation_1999-2021.xlsx"
	lookup_NUTSupdate = pd.read_excel(NUTSfile, sheet_name="All", header=0, index_col=None, 
		usecols=["Code_OLD","Code_NEW","Change"])
		
	## Define which types of changes can be simply replaced
	change_list = ["code change","recoded","code change, name changes","code change, label change","recoded and relabelled","recoded (boundary change)","name change","renamed"]
	updates = lookup_NUTSupdate.loc[lookup_NUTSupdate["Change"].str.lower().isin(change_list)]
	del lookup_NUTSupdate
	
	## Replace old area IDs with new
	change_dict = dict(zip(updates["Code_OLD"],updates["Code_NEW"]))
	data["Area ID"] = data["Area ID"].replace(change_dict, inplace=False)
	
	## Output
	if verbose: print(me+": Recoded NUTS2 regions old->new for panel data.")
	
	## Record list of updates
	if not nosave:
		outfile = "./Data/Diagnostic/NUTSUpdate1_recode.csv"
		updates.to_csv(outfile)
		if verbose: print(me+": NUTS recoding diagnostic report {:s}.".format(outfile))
	
	return data

##=================================================================================================

def clean_NUTSupdate_merge(data_yield, verbose, nosave):
	"""
	Replace old NUTS2 area codes with new ones
	This is the second (and more complex) part of a two-stage process of area code updates
	Here, we look at regions which have split or shifted, and back-fill their records with parent yields
	Specifically, we operate on "new region" and "boundary shift" changes
	
	Fill in yield data from split regions with parents
	"""
	
	## Read in file of ID changes
	NUTSfile = "./CodesLookup/NUTS2_ChangeCompilation_1999-2021.xlsx"
	lookup_NUTSupdate = pd.read_excel(NUTSfile,sheet_name="YieldMerge",header=0,index_col=None,
		usecols=["Code_OLD","Code_NEW"])
	
	## Identify rows which require back-filling -- i.e. are contained in the update file
	## Only consider rows whose names might have changed according to the records
	idx_fillto = data_yield.index.get_level_values("Area ID").isin(lookup_NUTSupdate["Code_NEW"].unique())
	
	## Create dictionary new->old for NUTS2 codes
	fill_dict = dict(zip(lookup_NUTSupdate["Code_NEW"],lookup_NUTSupdate["Code_OLD"]))
	del lookup_NUTSupdate
	
	## Loop through rows which may need filling
	## An expert would be able to do this faster, without the loop
	droplist = []
	for idx in data_yield.loc[idx_fillto,:].index:
		
		## Map the new area ID to the old ID; keep crop the same
		idx_fill = (idx[0],fill_dict[idx[1]])
		
		try:
			## Fill blanks in data_yield with values from the old area, if they exist
			data_yield.loc[idx,:].fillna(data_yield.loc[idx_fill,:],inplace=True)
			
			## Keep track of which indexes have been incorporated into rows for new areas
			droplist += [idx_fill]
			
		## If the old crop-area combination doesn't exist, move to the next
		except KeyError:
			pass

	## Output
	if verbose: print(me+": Back-filled old NUTS2 yield data for {:d} merged regions.".format(len(droplist)))
	
	## Record list of updates
	if not nosave:
		outfile = "./Data/Diagnostic/NUTSUpdate2_merge.csv"
		data_yield.loc[droplist,:].to_csv(outfile)
		if verbose: print(me+": NUTS merged and dropped rows report saved to {:s}.".format(outfile))
	
	## Delete any rows which have been incorporated
	data_yield = data_yield.drop(index=droplist, inplace=False)
	
	return data_yield

##=================================================================================================

def clean_dopivoting(data):
	"""
	Convert panel data into a pivoted table
	Returns a multi-indexed DataFrame
	"""
	
	## Set the columns to use as an index (extras will be popped into columns later)
	index_cols = ["Variable ID","Crop ID","Area ID"]
	extra_cols = [s.replace("ID","Name") for s in index_cols]

	## Pivot data so years are in columns
	## aggfunc combines identical rows
	data_pivot = pd.pivot_table(data, values=None, index=index_cols+extra_cols, columns=["Year"], aggfunc="sum")
	
	## Convert the extra part of the index into columns
	data_pivot.reset_index(level=extra_cols, inplace=True)
	
	## Set the precision
	data_pivot = data_pivot.round(4)
	
	return data_pivot

##=================================================================================================

def clean_duplicates_raw(data, datafile, verbose, nosave):
	"""
    Works on the unpivoted ("raw") dataset
    Outputs a file indicating duplicated data lines, for follow-up
	"""
	
	outdir = os.path.dirname(datafile)+"/Diagnostic/"
	
	## Check for duplicated entries in the data -- these will be summed in the pivot
	## Result is a 1D Series of bools; keep=False means both rows saved for comparison
	duplicates = data.duplicated(subset=["Variable ID","Area ID","Crop ID","Year"], keep=False)
	if duplicates.sum()>0 and not nosave:
		outfile = outdir+os.path.splitext(os.path.basename(datafile))[0]+"_diagnostic_duplicate_rawrows.csv"
		data.loc[duplicates,:].to_csv(outfile)
		print(me+": Detected {total:d} duplicate entries; saved to {outfile:s}".format(total=duplicates.sum(),outfile=outfile))

	return data

##=================================================================================================
	
def clean_duplicates_pivot(data_pivot, datafile, verbose, nosave):
	"""
    Works on the pivoted dataset
	Routines for eliminating duplicates from the pivoted data
	"""
	
	outdir = os.path.dirname(datafile)+"/Diagnostic/"
		
	## Checks for duplicated data rows, and outputs a file for follow-up
	## Result is a 1D Series of bools; keep=False means both rows saved for comparison
	duplicates = data_pivot.duplicated(subset=None, keep=False)
	if duplicates.sum()>0 and not nosave:
		outfile = outdir+os.path.splitext(os.path.basename(datafile))[0]+"_diagnostic_duplicate_pivotrows.csv"
		data_pivot.loc[duplicates,:].to_csv(outfile)
		print(me+": Detected {total:d} duplicate rows to drop; recorded to {outfile:s} for reference.".format(total=duplicates.sum(),outfile=outfile))

	## Drop exact (structural) duplicates -- keep the first entry and drop the second
	idx = data_pivot.index[data_pivot.duplicated(subset=None, keep="first")]
	data_pivot.drop(idx, axis=0, inplace=True)
	if verbose: print(me+": Detected {total:d} exact duplicate rows to drop.".format(total=duplicates.sum()))
	
	## Check for duplicates in the values, narrowing down to places that may have been renamed but have the same data
	duplicates = data_pivot["Value"].duplicated(subset=None, keep=False)
	if duplicates.sum()>0 and not nosave:
		outfile = outdir+os.path.splitext(os.path.basename(datafile))[0]+"_diagnostic_duplicate_pivotdata.csv"
		data_pivot.loc[duplicates,:].to_csv(outfile)
		if verbose: print(me+": Detected {total:d} duplicate data rows to drop; recorded to {outfile:s} for reference.".format(total=duplicates.sum(),outfile=outfile))

	## Drop rows where data is duplicated under a different name, but the Area ID is the same with an extra 0
	## Note that this throws away the second (alternative) name of the area
	duplicates = data_pivot["Value"].duplicated(subset=None, keep="first")
	idx = duplicates.index[duplicates.index.get_level_values("Area ID").str[-1]=="0"]
	data_pivot.drop(idx, axis=0, inplace=True)
	if verbose: print(me+": Detected {total:d} duplicate data rows with different Area IDs to drop.".format(total=duplicates.sum()))
	
	return data_pivot
	
##=================================================================================================
	
def clean_yielddata_maskjumps(data_yield, verbose, nosave):
	"""
	Find and mask jumps in the yield data where we decide there is an error
	The divergence threshold is a bit rudimentary; might want something more sophisticated
	"""
	
	##---------------------------------------------------------------------------------------------
	## Find value jumps
	
	## Find jumps, where value moves more than some multiple
	jump_factor = data_yield.div(np.roll(data_yield.values,1,axis=1), axis=1)
	jump_bool = (jump_factor > 2.0) | (jump_factor < 0.5)
	
	## Store row indexes
	idx_jump = data_yield.index[jump_bool.any(skipna=True, axis=1)]
	
	## Store only-jump df
	data_jump = data_yield.loc[idx_jump]

	## Produce a report of which rows jump for follow-up
	if not nosave:
		outfile = "./Data/Diagnostic/jumps.csv"
		data_jump.to_csv(outfile)
		if verbose: print(me+": Detected {:d} jumps; diagnostic report {:s}.".format(idx_jump.shape[0], outfile))
		
	##---------------------------------------------------------------------------------------------
	## Identify the more credible data segment in the jump rows
	
	## Calculate reference data: annual yield at the country-crop level and at the crop level (average over all countries)
	idx = data_yield.index.get_level_values("Area ID").str.len()==2
	data_country_crop = data_yield.loc[idx].rename_axis(index={"Area ID":"Country ID"},inplace=False)
	data_country_crop.loc[jump_bool.any(skipna=True, axis=1),:] = np.nan
	data_crop = data_country_crop.groupby(level="Crop ID").mean(skipna=True)
	
	## Create a df of reference data which is compatible with data_yield
	## Laborious: keep changing index so that fillna will work, and finally get it back into original form
	data_ref = pd.DataFrame(index=data_yield.index, columns=data_yield.columns, dtype=float)
	data_ref["Country ID"] = data_ref.index.get_level_values("Area ID").str[:2].values
	data_ref = data_ref.set_index("Country ID",append=True).reset_index(level="Area ID",drop=False).fillna(data_country_crop,inplace=False)
	data_ref = data_ref.reset_index(level="Country ID",drop=False,inplace=False).fillna(data_crop,inplace=False)
	data_ref = data_ref.set_index("Area ID",append=True,inplace=False).drop("Country ID",axis=1,inplace=False)
	
	## Calculate the divergence of the yield from the reference in the jump-rows only
	divergence = data_jump/data_ref.loc[idx_jump,:]
	
	##---------------------------------------------------------------------------------------------
	## Mask data

	## Mask values which cross a divergence threshold
	mask_bool = ((divergence > 2.5) | (divergence < 0.4))
	data_yield.where(~mask_bool,other=np.nan,inplace=True)
	
	##---------------------------------------------------------------------------------------------
	## Output	
	
	return data_yield
	
##=================================================================================================
	
def clean_yielddata_fillcrops(data_yield, verbose):
	"""
	Fill missing yield values with values from parent crop
	This cascades: a child receives parents' values, and a grandchild will then receive child's values
	
	There may be a more streamlined way to do this; however, it's not straightforward as parents have many children
	"""
	
	## Create a sorted list of crops
	crop_list = data_yield.sort_index(axis="index",level="Crop ID",inplace=False).index.get_level_values(\
		"Crop ID").unique().tolist()
	
	## Lookup table for crop IDs and associated parent crop IDs
	lookup_crop = read_CropID()
	parent_list = []
	
	## Loop over crops and replace missing values with their parents
	## An expert would be able to do this more efficiently without the loop
	for crop in crop_list:
		
		## Identify parent for the crop
		crop_parent = lookup_crop.loc[crop, "Parent ID"]
		
		## No need to do the fill if the crop is its own parent
		if crop_parent==crop: continue
		parent_list += [crop_parent]

		## Indexes (bool) of the yield data rows where the crop and the parent live
		idx_crop = data_yield.index.get_level_values("Crop ID") == crop
		idx_parent = data_yield.index.get_level_values("Crop ID") == crop_parent
		
		## Create a df with just the parent crop's values, using idx_parent
		## This will fill the gaps in the crop's dataset
		## Rename the parent crop to the crop so that the parent df is compatible
		data_yield_parent = data_yield.loc[idx_parent,:].rename({crop_parent:crop},
			axis="index",level="Crop ID",inplace=False)
		
		## Fill gaps (NAN) for the crop with values for the parent, which is using the same index
		## Must be done in-place
		data_yield.fillna(data_yield_parent, inplace=True)
	
	if verbose: print(me+": Filled missing values for each crop with country-specific values from the parent crop category.")
	
	return data_yield
	
##=================================================================================================

def check_rounding(data_pivot, datafile, verbose, nosave):
	"""
	Output a file that highlights where input data may have been rounded / made up
	No change implemented on the data: just outputs a reference file
	"""
	
    ## Define a output directory
	outdir = os.path.dirname(datafile)+"/Diagnostic/"
	
	## Rounded values: look for two consecutive years
	roundeds = (data_pivot["Value"].mod(1)==0)
	roundeds = roundeds & np.roll(roundeds.values,1,axis=1)		### Change this -- don't use roll
	
	## Save rows which appear to be rounded
	if roundeds.sum(axis=1).any() and not nosave:
		outfile = outdir+os.path.splitext(os.path.basename(datafile))[0]+"_diagnostic_pivot_roundedvalues.csv"
		data_pivot.loc[roundeds.any(axis=1),:].to_csv(outfile)
		print(me+": Detected {total:d} rounded entries on {rows:d} rows; output to {outfile:s} for follow-up. No changes made to dataset.".format(
			total=roundeds.sum().sum(),rows=roundeds.any(axis=1).sum(),outfile=outfile))
	
	return

##=================================================================================================

def clean_IDToName(df, lookup):
	"""
	General routine to map an ID to a long-form name and include in the input DataFrame
	Both df and lookup are DataFrames
	Lookup must have an ID index and a corresponding name column
	df must have the same ID column; a new name column will be created if necessary
	"""
	
	## New and old columns -- must be shared between df and lookup
	col_ID = lookup.index.name
	col_name = col_ID.replace("ID", "Name")
	
	## Create new column in df by converting ID column values to name values
	df[col_name] = df[col_ID].replace(
		to_replace=lookup.index.values, value=lookup[col_name].values,
		regex=False, inplace=False)
	
	return df

##=================================================================================================
##=================================================================================================
if __name__=="__main__":
	main()